@@ -19,11 +19,7 @@
 --
 ---------------------------------------------------------
 
--- | A fast, light-weight HTTP server handler for WAI. Some random notes (a FAQ, if you will):
---
--- * When a 'ResponseFile' indicates a file which does not exist, an exception
---   is thrown. This will close the connection to the client as well. You should
---   handle file existance checks at the application level.
+-- | A fast, light-weight HTTP server handler for WAI.
 module Network.Wai.Handler.Warp
     ( -- * Run a Warp server
       run
@@ -60,7 +56,6 @@ module Network.Wai.Handler.Warp
     , T.initialize
 #if TEST
     , takeHeaders
-    , readInt
 #endif
     ) where
 
@@ -86,6 +81,7 @@ import Control.Exception
     , fromException, AsyncException (ThreadKilled)
     , bracketOnError
     , IOException
+    , try
     )
 import Control.Concurrent (forkIO)
 import Data.Maybe (fromMaybe, isJust)
@@ -115,10 +111,10 @@ import Data.List (foldl')
 import Control.Monad (forever, when)
 import qualified Network.HTTP.Types as H
 import qualified Data.CaseInsensitive as CI
-import System.IO (hPutStrLn, stderr)
-import ReadInt (readInt64)
+import System.IO (hPrint, stderr)
 import qualified Data.IORef as I
 import Data.String (IsString (..))
+import qualified Data.ByteString.Lex.Integral as LI
 
 #if WINDOWS
 import Control.Concurrent (threadDelay)
@@ -198,7 +194,7 @@ bindPort p s = do
         theBody addr = 
           bracketOnError
           (Network.Socket.socket (addrFamily addr) (addrSocketType addr) (addrProtocol addr))
-          (sClose)
+          sClose
           (\sock -> do
               setSocketOption sock ReuseAddr 1
               bindSocket sock (addrAddress addr)
@@ -353,7 +349,7 @@ parseRequest' conn port (firstLine:otherLines) remoteHost' src = do
     let len0 =
             case lookup "content-length" heads of
                 Nothing -> 0
-                Just bs -> readInt bs
+                Just bs -> LI.readDecimal_ bs
     let serverName' = takeUntil 58 host -- ':'
     rbody <-
         if len0 == 0
@@ -511,29 +507,36 @@ sendResponse th req conn r = sendResponse' r
     needsChunked hs = isChunked' && not (hasLength hs)
     isKeepAlive hs = isPersist && (isChunked' || hasLength hs)
     hasLength hs = isJust $ lookup "content-length" hs
+    sendHeader = connSendMany conn . L.toChunks . toLazyByteString
 
     sendResponse' :: Response -> ResourceT IO Bool
-    sendResponse' (ResponseFile s hs fp mpart) = liftIO $ do
-        (lengthyHeaders, cl) <-
-            case (readInt `fmap` lookup "content-length" hs, mpart) of
-                (Just cl, _) -> return (hs, cl)
-                (Nothing, Nothing) -> do
+    sendResponse' (ResponseFile s hs fp mpart) = do
+        eres <-
+            case (LI.readDecimal_ `fmap` lookup "content-length" hs, mpart) of
+                (Just cl, _) -> return $ Right (hs, cl)
+                (Nothing, Nothing) -> liftIO $ try $ do
                     cl <- P.fileSize `fmap` P.getFileStatus fp
                     return $ addClToHeaders cl
                 (Nothing, Just part) -> do
                     let cl = filePartByteCount part
-                    return $ addClToHeaders cl
-        connSendMany conn $ L.toChunks $ toLazyByteString $
-          headers version s lengthyHeaders False
-
-        T.tickle th
-
-        if not (hasBody s req) then return isPersist else do
-            case mpart of
-                Nothing   -> connSendFile conn fp 0 cl (T.tickle th)
-                Just part -> connSendFile conn fp (filePartOffset part) (filePartByteCount part) (T.tickle th)
-            T.tickle th
-            return isPersist
+                    return $ Right $ addClToHeaders cl
+        case eres of
+            Left (_ :: SomeException) -> sendResponse' $ responseLBS
+                H.status404
+                [("Content-Type", "text/plain")]
+                "File not found"
+            Right (lengthyHeaders, cl) -> liftIO $ do
+                let headers' = headers version s lengthyHeaders
+                sendHeader $ headers' False
+                T.tickle th
+                if hasBody s req then do
+                    case mpart of
+                        Nothing   -> connSendFile conn fp 0 cl (T.tickle th)
+                        Just part -> connSendFile conn fp (filePartOffset part) (filePartByteCount part) (T.tickle th)
+                    T.tickle th
+                    return isPersist
+                  else
+                    return isPersist
       where
         addClToHeaders cl = (("Content-Length", B.pack $ show cl):hs, fromIntegral cl)
 
@@ -544,10 +547,7 @@ sendResponse th req conn r = sendResponse' r
                 T.tickle th) body
               return (isKeepAlive hs)
         | otherwise = liftIO $ do
-            connSendMany conn
-                $ L.toChunks
-                $ toLazyByteString
-                $ headers' False
+            sendHeader $ headers' False
             T.tickle th
             return isPersist
       where
@@ -559,8 +559,16 @@ sendResponse th req conn r = sendResponse' r
                        `mappend` chunkedTransferTerminator
                   else headers' False `mappend` b
 
-    sendResponse' (ResponseSource s hs bodyFlush) =
-        response
+    sendResponse' (ResponseSource s hs bodyFlush)
+        | hasBody s req = do
+            let src = CL.sourceList [headers' needsChunked'] `mappend`
+                      (if needsChunked' then body C.$= chunk else body)
+            src C.$$ builderToByteString C.=$ connSink conn th
+            return $ isKeepAlive hs
+        | otherwise = liftIO $ do
+            sendHeader $ headers' False
+            T.tickle th
+            return isPersist
       where
         body = fmap (\x -> case x of
                         C.Flush -> flush
@@ -568,26 +576,12 @@ sendResponse th req conn r = sendResponse' r
         headers' = headers version s hs
         -- FIXME perhaps alloca a buffer per thread and reuse that in all
         -- functions below. Should lessen greatly the GC burden (I hope)
-        response
-            | not (hasBody s req) = liftIO $ do
-                connSendMany conn
-                   $ L.toChunks $ toLazyByteString
-                   $ headers' False
-                T.tickle th
-                return (checkPersist req)
-            | otherwise = do
-                let src =
-                        CL.sourceList [headers' needsChunked'] `mappend`
-                        (if needsChunked' then body C.$= chunk else body)
-                src C.$$ builderToByteString C.=$ connSink conn th
-                return $ isKeepAlive hs
         needsChunked' = needsChunked hs
         chunk :: C.Conduit Builder IO Builder
         chunk = C.Conduit
             { C.conduitPush = push
             , C.conduitClose = close
             }
-
         conduit = C.Conduit push close
         push x = return $ C.Producing conduit [chunkedTransferEncoding x]
         close = return [chunkedTransferTerminator]
@@ -695,7 +689,7 @@ defaultSettings = Settings
             Just x -> go x
             Nothing ->
                 when (go' $ fromException e) $
-                    hPutStrLn stderr $ show e
+                    hPrint stderr e
     , settingsTimeout = 30
     , settingsIntercept = const Nothing
     , settingsManager = Nothing
@@ -769,11 +763,6 @@ checkCR bs pos =
         else pos
 {-# INLINE checkCR #-}
 
-readInt :: Integral a => ByteString -> a
-readInt bs = fromIntegral $ readInt64 bs
-{-# INLINE readInt #-}
-
-
 -- | Call the inner function with a timeout manager.
 withManager :: Int -- ^ timeout in microseconds
             -> (Manager -> IO a)
